#include "ipv4.h"
#include <limits.h>

#define IP_STRLEN 16

#ifdef COLORS
#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"
#else
#define ANSI_COLOR_RED     ""
#define ANSI_COLOR_GREEN   ""
#define ANSI_COLOR_YELLOW  ""
#define ANSI_COLOR_BLUE    ""
#define ANSI_COLOR_MAGENTA ""
#define ANSI_COLOR_CYAN    ""
#define ANSI_COLOR_RESET   ""
#endif

unsigned char * iptouca
(const uint32_t ip, unsigned char * dst, const int length)
{
  if (length != 4)
    {
      fprintf(stderr, ANSI_COLOR_RED "Cannot convert integer IP to array in %d bytes."
	      ANSI_COLOR_RESET, length);
      return NULL;
    }
  else
    {
      memset(dst, 0, sizeof(unsigned char) * 4);
    }

  dst[0] = (unsigned char)(ip >> 24);
  dst[1] = (unsigned char)(ip >> 16);
  dst[2] = (unsigned char)(ip >>  8);
  dst[3] = (unsigned char)(ip & 0x000000ff);

  return dst;
}

char * ipntoa
(const uint32_t ip, char * dst, const int length)
{
  unsigned char oct[4];

  if (length < IP_STRLEN)
    {
      fprintf(stderr, "ipntoa: buffer must be at least %d bytes wide.\n",
		IP_STRLEN);
      exit(-1);
    }

  if (dst == NULL)
    {
      dst = malloc(sizeof(char) * IP_STRLEN);
    }

  memset(dst, 0, length); 
  if (iptouca(ip, oct, 4 * sizeof(char)) == NULL)
    {
      exit(EXIT_FAILURE);
    }

  if (snprintf(dst, length, "%d.%d.%d.%d", oct[0], oct[1], oct[2], oct[3]) == length)
    {
      fprintf(stderr, "Warning: ipntoa exceeded buffer size.\n");
    }
  return(dst);
}

uint32_t networkNumber
(uint32_t ip, uint32_t sm)
{
    return (ip & sm);
}

uint32_t subnetDirectedBroadcast
(route r)
{
    return (r.ip | ~r.sm);
}

uint32_t wildcardMask
(uint32_t sm)
{
    return ~sm;
}

char classful
(const uint32_t ip)
{
    char class = 0;
    if (ip >= 0xf0000000)      /* High bits = 1111 */
        class = 'E';
    else if (ip >= 0xe0000000) /* High bits = 1110 */
        class = 'D';
    else if (ip >= 0xc0000000) /* High bits = 110  */
        class = 'C';
    else if (ip >= 0x80000000) /* High bits = 10   */
        class = 'B';
    else                       /* High bit  = 0    */
        class = 'A';
    return(class);
}

uint32_t classfulMask
(const uint32_t ip)
{
    uint32_t mask = 0;
    if (ip >= 0xc0000000)       /* High bits = 110 */
        mask = 0xffffff00;
    else if (ip >=  0x80000000) /* High bits = 10  */
        mask = 0xffff0000;
    else                        /* High bit  = 0   */
        mask = 0xff000000;
    return(mask);
}

uint32_t prefixLength
(const uint32_t sm)
{
  return ((33 - __builtin_ffs(sm)) % 33); /* mod 33 to handle special 0 case */
}

uint32_t countHosts
(const uint32_t sm)
{
  uint32_t prefix, hosts;
  prefix = prefixLength(sm);
  if (sm > 0x00000000 && sm < 0xfffffffe) /* 1.0.0.0 - 255.255.255.252 */
    {
      hosts = (1 << (32 - prefix)) - 2;
    }
  /* The /31 network isn't often used in IPv4. This is a corner case. */
  else if (sm == 0xfffffffe)              /* 255.255.255.254 */
    {
      hosts = 2;
    }
  else                                    /* 0.0.0.0 and 255.255.255.255 */
    {
      hosts = 0;
    }
  return(hosts);
}

uint32_t countSubnets
(const uint32_t ip, const uint32_t sm)
{
    uint32_t subnets = 0;
    uint32_t classful = classfulMask(ip);

    if (sm > classful)
    {
        subnets = (1 << (prefixLength(sm) - prefixLength(classful)));
    }

    return(subnets);
}

int isValidNetmask
(const uint32_t sm)
{
  uint32_t valid;
  int trailz, popcount;

  valid = INT_MAX;

  /* The number of training zeroes after the first field set. */
  trailz = __builtin_ffs(sm) - 1;

  /* Population Count: the number of active bits in integer. */
  popcount = __builtin_popcount(sm);

  if (sm == 0)
    {
      fprintf(stderr, ANSI_COLOR_CYAN
	      "Warning: 0.0.0.0 subnet mask.\n"
	      ANSI_COLOR_RESET);
      valid = 1;
    }
  else if (sm == 0xffffffff)
    {
      fprintf(stderr, ANSI_COLOR_CYAN
	      "Warning: 255.255.255.255 subnet mask.\n"
	      ANSI_COLOR_RESET);
      valid = 1;
    }
  /* A valid subnet mask will have all of the zeroes at the end
     (eg: 11111111.11111111.11111111.11110000). This means that index
     of the first field set should be equal to the number of zeros.
     Using this test, we can detect invalid subnet masks that contain
     an out-of-place zero.
     For example: 11111111.11111110.11111111.11111000 has a population
     count of 28, but the First Field Set is 4.
     32 - (28 + (4-1)) > 0, so this is not a valid subnet mask. */
  else if (trailz + popcount == 32)
    {
      valid = 1;
    }
  else
    {
      char str[16];
      fprintf(stderr,
	      ANSI_COLOR_RED
	      "%s is not a valid subnet mask (FFS=%d, POP=%d).\n"
	      ANSI_COLOR_RESET,
	      ipntoa(sm, str, IP_STRLEN), trailz, popcount);
      valid = 0;
    }
  return (valid);
}

uint32_t parse
(const char * src)
{
  uint32_t ip, i, x;
  char * token;
  char s[IP_STRLEN];
  const char delim[2] = ".";

  strncpy(s, src, IP_STRLEN);

  if (strlen(s) > IP_STRLEN)
    {
      fprintf(stderr, ANSI_COLOR_RED "%s is not a valid IP.\n"
	      ANSI_COLOR_RESET, s);
      exit(-1);
    }
  else /* check for characters other than 0-9 and '.' */
    {
      int a;
      for (a = 0 ; a < (char) strlen(src) ; a++)
	{
	  if (s[a] != '.' && (s[a] < '0' || s[a] > '9'))
	    {
	      fprintf(stderr, ANSI_COLOR_RED
		      "%s is not a valid IP (invalid character: %c).\n"
		      ANSI_COLOR_RESET, src, s[a]);
	      exit(-1);
	    }
	}
    }

  ip = 0;
  i = 32;

  token = strtok(s, delim);
  while (i > 0 && token != NULL)
    {
      i -= 8;
      x = (uint32_t) atoi(token);
      
      if (x > 255)
	{
	  fprintf(stderr, ANSI_COLOR_RED
		  "%s is not a valid IP (invalid value: %s)\n"
		  ANSI_COLOR_RESET, src, token);
	  exit(-1);
	} 

      ip += (x << i);
      token = strtok(NULL, delim);
    }
  return(ip);
}

void generateSubnets
()
{
  uint32_t mask, i;
  char ip[IP_STRLEN];
  mask = 0x00000000;
  for (i=0; i<33; i++) /* though ugly, 33 makes sure we catch case 0xffffffff */
    {
      ipntoa(mask, ip, IP_STRLEN);
      printf("%-2d\t%s\n", prefixLength(mask), ip);
      mask = (mask >> 1) + 0x80000000;
    }
}

uint32_t summaryAddress
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
  return (networkNumber(ip1, sm1) & networkNumber(ip2, sm2) 
	  & summaryPrefix(ip1, sm1, ip2, sm2));
}

uint32_t summaryPrefix
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
  /* The Common Network Mask is the bits shared by each IP network. */
  uint32_t cnm;

  /* This is the Prefix that summarizes the two networks.
     It can be equal to or shorter than either input subnet mask. */
  uint32_t pfx;

  /* These are the bits different in each network number. */
  uint32_t diff;

  uint32_t network1, network2;

#ifdef DEBUG_SUMMARY_PREFIX
  char diff_s [IP_STRLEN];
  char cnm_s [IP_STRLEN];
  char pfx_s [IP_STRLEN];
  char sm_s [IP_STRLEN];
#endif

  network1 = networkNumber(ip1, sm1);
  network2 = networkNumber(ip2, sm2);

  /* Extra & against opposite subnet mask discovers
      if one network is a subnet of the other,
      or the same network. */
  if ((network1 & sm2) == (network2 & sm1))
    {
      pfx = sm1 & sm2;
    }
  else
    {
      diff = network1 ^ network2;

      /* Possible bugs here - clz can produce unexpected results.
	 The problem is if the CLZ instruction is issued to the
	 CPU but not supported. Intel will not cause an error,
	 and this algorithm will enter an undefined state. */
      cnm = 0xffffffff << (32 - __builtin_clz(diff));

      if (cnm > (sm1 & sm2))
	{
	  fprintf(stderr, ANSI_COLOR_RED);
	  fprintf(stderr,
		  "QA Warning: this line should not be mathematically"
		  " unreachable. Please report a bug in summary prefix.\n");
	  fprintf(stderr, ANSI_COLOR_RESET);
	  pfx = (sm1 & sm2);
	}
      else
	{
	  pfx = cnm;
	}
    }

#ifdef DEBUG_SUMMARY_PREFIX
  fprintf(stderr, ANSI_COLOR_BLUE);
  fprintf(stderr, "                 SUMMARY PREFIX\n");
  fprintf(stderr, "diff = %-16s\tclz(diff) = %d\n",
	  ipntoa(diff, diff_s, IP_STRLEN), __builtin_clz(diff));
  fprintf(stderr, "cnm  = %-16s\tsm1 & sm2 = %s\n",
	  ipntoa(cnm, cnm_s, IP_STRLEN), ipntoa(sm1 & sm2, sm_s, IP_STRLEN));
  fprintf(stderr, "pfx  = %-16s\n", ipntoa(pfx, pfx_s, IP_STRLEN));
  fprintf(stderr, ANSI_COLOR_RESET "\n");
#endif

  return pfx;
}

route * summaryClassful
(route r1)
{
  static route * classfulRoutes;
  int numRoutes;

  numRoutes = summaryClassfulCount(r1);

  /* The prefix is too short to fit into one classful
     route, so we calculate all the classful routes
     required for this supernet. */
  if (numRoutes > 1)
    {
      int i, increment;
      uint32_t sm = classfulMask(r1.ip);

      classfulRoutes = calloc(numRoutes, sizeof(route));

      increment = countHosts(sm) + 2;

      for (i=0; i<numRoutes; i++)
	{
	  classfulRoutes[i].ip = (r1.ip + (i * increment));
	  classfulRoutes[i].sm = sm;
	}
    }
  /* The route was either already at the classful boundary,
     or the prefix is longer than the classful boundary. */
  else 
    {
      route cr;

      numRoutes = 1;
      classfulRoutes = calloc(numRoutes, sizeof(route));
      cr.sm = classfulMask(r1.ip);
      cr.ip = r1.ip & cr.sm;
      classfulRoutes[0] = cr;
    }

  return classfulRoutes;
}

int summaryClassfulCount
(route r1)
{
  /* Classful Subnet Mask Population Count */  
  int csmpop;

  /* Summary Subnet Mask Population Count */
  int ssmpop;

  /* The number of classful networks contained in this summary */
  int count;

  csmpop = __builtin_popcount(classfulMask(r1.ip));
  ssmpop = __builtin_popcount(r1.sm);

#ifdef DEBUG
  fprintf(stderr, "Classful Subnet Mask Population Count: %d\n", csmpop);
  fprintf(stderr, "Original Subnet Mask Population Count: %d\n", ssmpop);
#endif

  count = (csmpop > ssmpop) ? 1 << (csmpop - ssmpop) : 1;

  return count;
}
